/**
 * HinzufuegenDialog.java
 * eu.gronos.kostenrechner (Kostenrechner)
 */
package eu.gronos.kostenrechner.view;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagLayout;
import java.awt.HeadlessException;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.DecimalFormat;
import java.text.NumberFormat;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JEditorPane;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRootPane;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.text.NumberFormatter;

import eu.gronos.kostenrechner.Kostenrechner;
import eu.gronos.kostenrechner.view.helpanderror.FehlerDialog;

/**
 * Oberklasse für alle Hinzufügen-Dialoge
 * 
 * @author Peter Schuster (setrok)
 * @date: 29.06.2014
 * 
 */
public abstract class HinzufuegenDialog<E> extends JDialog {

	/**
	 * Der long mit dem Namen serialVersionUID speichert...
	 */
	private static final long serialVersionUID = -6595384338088840389L;
	public static final String FEHLER_BT_UNTERLIEGEN_GT_STREITWERT = "Der Beteiligte kann nicht zu mehr verurteilt werden, als von ihm eingeklagt wurde!";
	public static final String FEHLER_WK_STREITWERT = "Bei einer Widerklage muss der Streitwert > 0 sein.";
	public static final String FEHLER_ZAHLFORMAT_DWBK = "Ungültiges Zahlformat bei der Anzahl der Drittwiderbeklagten";
	public static final String FEHLER_NIX_WIDERKLAEGER_WIDERBEKLAGTER = "Kein widerklagender Beklagter oder kein Widerbeklagter asugewählt!";
	public static final String FEHLER_UNTERLIEGEN_GT_STREITWERT = "Das Unterliegen darf nicht höher als der Streitwert sein! Bei Gesamtschuldnern wird das gesamtschuldnerische Unterliegen mit berücksichtigt.";
	public static final String FEHLER_ZAHLENFORMAT = "Fehler beim Konvertieren des Zahlenformats!";
	public static final String FEHLER_BK_GESAMTSCHULDNERSCHAFT = "Wenn Sie gesamtschuldnerische Verurteilung aktivieren, muss mehr als ein Beklagter markiert sein, die Höhe der Verurteilung > 0 sein, aber maximal so hoch wie der Streitwert sein.";
	public static final String FEHLER_WK_GESAMTSCHULDNERSCHAFT = "Wenn Sie gesamtschuldnerische Verurteilung aktivieren, muss mehr als ein Widerbeklagter markiert sein, die Höhe der Verurteilung > 0 sein, aber maximal so hoch wie der Streitwert sein.";
	public static final String FEHLER_ZAHLENFORMAT_SW = "Falches Zahlformat beim Streitwert!";
	public static final String FEHLER_KLAEGER_NIX_WBK = "Der ausgewählte Kläger ist nicht widerbeklagt.";
	public static final String FEHLER_ZAHLENFORMAT_KL_BK = "Falsches Zahlformat bei Kläger- oder Beklagtenanzahl! ";
	public static final String FEHLER_BEIM_ANLEGEN_KL_BK = "Fehler beim Anlegen der Kläger oder Beklagten! ";
	public static final String FEHLER_AUSLAGENHOEHE_NEGATIV = "Die Auslagenhöhe darf nicht negativ sein!";
	public static final String FEHLER_ANZAHL_NEGATIV = "Die Anzahl darf nicht negativ sein!";
	public static final String FEHLER_ZAHLENFORMAT_WERT = "Falsches Eingabeformat beim Wert! ";
	public static final String FEHLER_BEGRUENDETHEIT_GT_WERT = " Die Begründetheit der Forderung kann nicht höher als der Wert sein!";
	/**
	 * Der EmptyBorder mit dem Namen BORDER speichert...
	 */
	protected static final EmptyBorder BORDER = new EmptyBorder(5, 5, 5, 5);
	/**
	 * Der JPanel mit dem Namen contentPanel ist das JPanel für den eigentlichen
	 * Inhalt
	 */
	protected final JPanel contentPanel = new JPanel();
	/**
	 * Der boolean mit dem Namen isOk speichert, ob der JDialog mit OK oder Cancel
	 * verlassen wurde.
	 */
	private boolean isOk = false;
	private JPanel buttonPane;
	/**
	 * Der NumberFormatter mit dem Namen formatter kann für Nummereingabefelder
	 * benutzt werden. Grouping ist ausgeschaltet, Eingaben können fehlerhaft sein.
	 */
	protected NumberFormatter formatter;
	private final AbstractAction okAction = new AbstractAction("OK") {
		private static final long serialVersionUID = -5819524266780763824L;

		{
			putValue(ACTION_COMMAND_KEY, "OK");
			putValue(MNEMONIC_KEY, KeyEvent.VK_O);
			putValue(SHORT_DESCRIPTION, "Bestätigt die Eingabe (Eingabetaste).");
			putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));
		}

		@Override
		public void actionPerformed(ActionEvent ae) {
			if (parseEingabe()) {
				isOk = true;
				setVisible(false);
				dispose();
			} else
				return;
		}
	};

	private final Action cancelAction = new AbstractAction("Abbrechen") {
		private static final long serialVersionUID = 7626690970693233716L;
		{
			putValue(ACTION_COMMAND_KEY, "Cancel");
			putValue(SHORT_DESCRIPTION, "Bricht die Eingabe ab (ESC).");
			putValue(MNEMONIC_KEY, KeyEvent.VK_A);
			putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
		}

		@Override
		public void actionPerformed(ActionEvent ae) {
			fensterSchliessenBeiAbbruch();
		}
	};

	/**
	 * fensterSchliesser soll dafür sorgen, dass sich das Fenster über die
	 * Schließen-Schaltfläche ("X" bzw. roter Button) verhält wie bei Abbruch
	 */
	private final WindowAdapter fensterSchliesser = new WindowAdapter() {

		@Override
		public void windowClosing(WindowEvent e) {
			fensterSchliessenBeiAbbruch();
		}

	};

	/**
	 * Der FocusAdapter für JTextField und davon abgeleitete markiert bei
	 * Focuserlangung den gesamten Inhalt.
	 * 
	 * JFormattedTextField muss immer, wenn es den Fokus bekommt, seinen Inhalt
	 * komplett markieren, damit er komplett überschrieben wird.
	 * 
	 * @url "http://www.java-forum.org/awt-swing-javafx-and-swt/20553-jtextfield-kompletten-text-selektion-markieren.html"
	 */
	public static final FocusAdapter BEI_FOCUS_MARKIEREN = new FocusAdapter() {

		/**
		 * Markiert bei Focuserlangung den gesamten Inhalt
		 * 
		 * @param fe ein FocusEvent
		 * 
		 * @see java.awt.event.FocusAdapter#focusGained(java.awt.event.FocusEvent)
		 */
		@Override
		public void focusGained(FocusEvent fe) {
			Object o = fe.getSource();
			if (o instanceof JFormattedTextField || o instanceof JTextField) {
				selectAll((JTextField) o);
			}
		}

		@Override
		public void focusLost(FocusEvent fe) {
			Object o = fe.getSource();
			if (o instanceof JFormattedTextField || o instanceof JTextField) {
				selectAll((JTextField) o);
			}
		}

		private void selectAll(final JTextField tf) {
			/* Muss eigener Runnable sein, sonst funktioniert das nicht. */
			SwingUtilities.invokeLater(new Runnable() {
				@Override
				public void run() {
					tf.selectAll();
				}
			});
		}

	};
	private JButton okButton;
	private boolean okButtonToBeFocussed = false;
	/**
	 * HYPERLINK_LISTENER ist ein HyperlinkListener, der auf Klicks reagiert und
	 * entweder interne Links im selben Dokument anspringt oder für externe Links
	 * den Standardbrowser öffnet.
	 * 
	 * @url "http://www.dpunkt.de/java/Programmieren_mit_Java/Oberflaechenprogrammierung/59.html"
	 * @url "http://stackoverflow.com/questions/10967451/open-a-link-in-browser-with-java-button"
	 * @url "http://stackoverflow.com/questions/14054601/jumping-to-internal-anchors-in-jeditorpane"
	 */
	public static final HyperlinkListener HYPERLINK_LISTENER = new HyperlinkListener() {

		final Desktop desktop = Desktop.isDesktopSupported() ? Desktop.getDesktop() : null;

		@Override
		public void hyperlinkUpdate(HyperlinkEvent hle) {
			if (hle.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
				if (!isNull(hle) && !interneSprungmarke(hle) && hle.getURL() != null)
					launchBrowser(hle.getURL());

			}
		}

		/**
		 * Die Methode dient dazu, alle denkbaren {@link NullPointerException}s
		 * auszuschließen
		 * 
		 * @param hle ein {@link HyperlinkEvent}
		 * @return <code>true</code>, wenn {@link Desktop#isDesktopSupported()}
		 *         <code>false</code> ist, {@link HyperlinkEvent#getURL()} oder
		 *         {@link HyperlinkEvent#getDescription()} <code>null</code> liefern.
		 */
		private boolean isNull(HyperlinkEvent hle) {
			if (hle.getDescription() == null) {
				// System.out.println
				Kostenrechner.getLogger().info("Keine Linkbeschreibung");
				return true;
			}
			return desktop == null;
		}

		/**
		 * Die Methode dient dazu, interne Sprungmarken anzusteuern
		 * 
		 * @param hle das {@link HyperlinkEvent}
		 * @return <code>true</code>, wenn es sich um eine Sprungmarke handelte.
		 */
		private boolean interneSprungmarke(HyperlinkEvent hle) {
			String desc = hle.getDescription();
			if (!desc.startsWith("#")) {
				/* Ohne Raute keine Sprünge */
				// System.out.println(
				Kostenrechner.getLogger().info(String.format("Externer Link: %s", desc));
				return false;
			} else {
				/* Raute (#) entfernen */
				// System.out.println(
				Kostenrechner.getLogger().info(String.format("Interner Link: %s", desc));
				desc = desc.substring(1);
			}
			if (hle.getSource() instanceof JEditorPane) {
				JEditorPane ep = (JEditorPane) hle.getSource();
				ep.scrollToReference(desc);
			} else
				// System.out.println
				Kostenrechner.getLogger().info("Kein Editorpane");
			return true;
		}

		/**
		 * Die Methode übernimmt das Konvertieren in eine {@link URI} und ruft über
		 * {@link #launchBrowser(URI)} den Standardbrowser des Anwenders auf.
		 * 
		 * @param url die anzusteuernde {@link URL}
		 */
		private void launchBrowser(URL url) {
			try {
				launchBrowser(url.toURI());
				// System.out.println(
				Kostenrechner.getLogger().info(String.format("URL: %s", url.toString()));
			} catch (URISyntaxException use) {
				FehlerDialog.zeigeFehler(FehlerDialog.FEHLER_BLANKO,
						"Der ausgewählte Link enthält einen Syntaxfehler!", use);
				use.printStackTrace();
			}
		}

		/**
		 * Die Methode ruft den Standardbrowser des Anwenders auf.
		 * 
		 * @param uri die anzuspringende {@link URI}
		 * 
		 * @see Desktop#browse(URI)
		 */
		private void launchBrowser(URI uri) {
			if (desktop != null && desktop.isSupported(Desktop.Action.BROWSE)) {
				try {
					desktop.browse(uri);
				} catch (IllegalArgumentException | NullPointerException iae) {
					FehlerDialog.zeigeFehler(FehlerDialog.FEHLER_BLANKO,
							"URI konnte nicht geöffnet werden oder Link ist NULL", iae);
				} catch (SecurityException se) {
					FehlerDialog.zeigeFehler(FehlerDialog.FEHLER_BLANKO,
							"Sicherheitsverletzung beim Start des Browsers.", se);
				} catch (IOException ioe) {
					FehlerDialog.zeigeFehler(FehlerDialog.FEHLER_BLANKO,
							"Browser konnte nicht gestartet werden. Kein Standardbrowser definiert, definierter Standardbrowser nicht gefunden oder das System konnte ihn nicht starten.",
							ioe);
				} catch (Exception e) {
					FehlerDialog.zeigeFehler(FehlerDialog.FEHLER_BLANKO, "Sonstiger Fehler beim Starten des Browsers",
							e);
				}
			}
		}
	};

	/**
	 * Erstellt einen modalen JDialog mit einer buttonPane (OK und Abbrechen) und
	 * einem contentPanel. Dieser muss von der abgeleiteten Klasse in der Methode
	 * {@link #fuelleContentPane()} mit JComponents gefüllt werden, die von
	 * {@link #showDialog()} aufgerufen wird.
	 * 
	 * @param owner the JFrame from which the dialog is displayed
	 * @param title Der anzuzeigende Titel
	 * @throws HeadlessException wenn in einer Umgebung aufgerufen, die keinen
	 *                           Bildschirm, Tastatur oder Maus unterstützt
	 */
	public HinzufuegenDialog(JFrame owner, String title) throws HeadlessException {
		super(owner, title, true);
		baueContentPane();
		setButtonPane();
		formatter = getNumberFormatter();
	}

	/**
	 * Die Methode dient dazu, aus den Eingabefeldern des Dialogs einen Rückgabewert
	 * zusammenzubauen, der dann von {@link #showDialog()} an die aufrufende Maske
	 * zurückgegeben werden kann.
	 * 
	 * @return ein <code>&lt;E&gt;</code> mit allen Werten
	 */
	protected abstract E baueRueckgabewert();

	/**
	 * Die Methode füllt den Inhaltsbereich des Dialogs mit JComponents zur Eingabe.
	 * Wird von {@link #showDialog()} aufgerufen.
	 * 
	 */
	protected abstract void fuelleContentPane();

	/**
	 * Die Methode wird von {@link #okAction} aufgerufen, um zu prüfen, ob alle
	 * JComponents zur Eingabe mit gültigen Werten gefüllt sind und diese zueinander
	 * nicht im Widerspruch stehen.
	 * 
	 * @return true, wenn die Plausibilitäten zufrieden sind
	 */
	protected abstract boolean parseEingabe();

	/**
	 * Die Methode initialisiert die allgemeinen JComponents für den Inhaltsbereich
	 * des Dialogs und wird vom Konstruktor selbsttätig aufgerufen.
	 * 
	 */
	protected void baueContentPane() {
		GridBagLayout gbl_contentPanel = new GridBagLayout();
		contentPanel.setBorder(BORDER);
		contentPanel.setLayout(gbl_contentPanel);
		getContentPane().add(contentPanel, BorderLayout.CENTER);
	}

	/**
	 * Die Methode baut einen Numberformatter
	 * 
	 * @return den NumberFormatter
	 */
	private NumberFormatter getNumberFormatter() {
		NumberFormat format = DecimalFormat.getInstance();
		// NumberFormat format = NumberFormat.getInstance();
		NumberFormatter formatter = new NumberFormatter(format);
		format.setGroupingUsed(false);
		formatter.setAllowsInvalid(true);
		return formatter;
	}

	/**
	 * Die Methode setzt die OK- und die Abbrechen-Schaltfläche in einen eigenen
	 * JPanel des Dialogs
	 * 
	 */
	private void setButtonPane() {
		buttonPane = new JPanel();
		buttonPane.setLayout(new FlowLayout(FlowLayout.RIGHT));
		getContentPane().add(buttonPane, BorderLayout.SOUTH);
		okButton = new JButton(okAction);
		addToRootInputMap(okAction);
		buttonPane.add(okButton);
		getRootPane().setDefaultButton(okButton);
		JButton cancelButton = new JButton(cancelAction);
		buttonPane.add(cancelButton);
		addToRootInputMap(cancelAction);
	}

	/**
	 * @return gibt den buttonPane als JPanel zurück.
	 */
	protected JPanel getButtonPane() {
		return buttonPane;
	}

	/**
	 * Die Methode fügt eine Action der ActionMap und der InputMap der JRootPane des
	 * HinzufuegenDialogs hinzu.
	 * 
	 * @param action eine Action, bei der die Values
	 *               {@link Action#ACTION_COMMAND_KEY} und
	 *               {@link Action#ACCELERATOR_KEY} gesetzt sein müssen.
	 */
	protected void addToRootInputMap(Action action) {
		JRootPane rp = getRootPane();
		// Zunächst die Action der ActionMap hinzufügen unter ihrem Namen
		rp.getActionMap().put(action.getValue(Action.ACTION_COMMAND_KEY), action);
		// Dann die Action der InputMap unter ihrer Tastenkombination hinzufügen
		rp.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put((KeyStroke) action.getValue(Action.ACCELERATOR_KEY),
				action.getValue(Action.ACTION_COMMAND_KEY));
		// rp.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(
		// KeyStroke.getKeyStroke(
		// (Integer) action.getValue(Action.ACCELERATOR_KEY),
		// InputEvent.ALT_DOWN_MASK),
		// action.getValue(Action.ACTION_COMMAND_KEY));
	}

	/**
	 * Die Methode showDialog dient dazu, den Dialog aufzurufen, den
	 * {@link #contentPanel} über die Methode {@link #fuelleContentPane()} der
	 * abgeleiteten Klasse mit Komponenten zu füllen, mit {@link #setBounds()} die
	 * Größe des Dialogs anzupassen und auf eine Eingabe zu warten und dann einen E
	 * zurückzugeben. Zur Zurückgabe der Eingabe wird {@link #baueRueckgabewert()}
	 * der abgeleiteten Klasse aufgerufen.
	 * 
	 * @return einen <code>&lt;E&gt;</code> nach den eingegebenen Werten
	 */
	public E showDialog() {
		fuelleContentPane();
		// setBounds();
		pack();
		setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		addWindowListener(fensterSchliesser);
		// OK-Schaltfläche fokussieren, wenn gewünscht
		if (okButtonToBeFocussed)
			okButton.requestFocus();
		setVisible(true);
		// Hier wartet der wohl...
		if (isOk) {
			return baueRueckgabewert();
		} else {
			return null;
		}
	}

	/**
	 * Die Methode getPreferredSize dient dazu, die bevorzugte Größe des JDialogs zu
	 * ermitteln, indem die Größe der Ränder hinzugerechnet wird.
	 * 
	 * @return eine {@link java.awt.Dimension} mit der vergrößerten bevorzugten
	 *         Größe
	 * 
	 * @see java.awt.Container#getPreferredSize()
	 * @see java.awt.Container#getMinimumSize()
	 */
	public Dimension getPreferredSize() {
		Dimension ps = new Dimension();
		Insets i = BORDER.getBorderInsets();
		double rand = (double) (i.left + i.right + i.top + i.bottom);
		// Liefert getMinimumSize() oder getPreferredSize() den größeren Wert
		// für Breite und Höhe?
		final Dimension gms = super.getMinimumSize();
		final Dimension gps = super.getPreferredSize();
		// Nimm die größere Breite
		double width = gms.getWidth() > gps.getWidth() ? gms.getWidth() : gps.getWidth();
		// Nimm die größere Höhe
		double height = gms.getHeight() > gps.getHeight() ? gms.getHeight() : gps.getHeight();
		// Rand hinzurechnen
		ps.setSize(width + rand, height + rand);
		return ps;
	}

	/**
	 * Die Methode lässt die Position vom Betriebssystem zuweisen und setzt die
	 * Größe auf die bevorzugte Größe. Wird von {@link #showDialog()} aufgerufen.
	 * 
	 * @see java.awt.Window#setBounds(int,int,int,int)
	 * @see java.awt.Window#setLocationByPlatform(boolean)
	 * @see java.awt.Window#setSize(Dimension)
	 * @see java.awt.Window#pack()
	 */
	@Override
	public void pack() {
		setLocationByPlatform(true);
		setSize(getPreferredSize());
		super.pack();
	}

	/**
	 * Ordnet ein JLabel einer JComponent zu und setzt den Mnemonic key.
	 * 
	 * @param label     ein JLabel
	 * @param component the Component this label is for, or null if the label is not
	 *                  the label for a component
	 * @param key       a keycode that indicates a mnemonic key
	 * 
	 * @see javax.swing.JLabel#setLabelFor(Component)
	 * @see javax.swing.JLabel#setDisplayedMnemonic(int)
	 */
	protected void setMnemonicLabelFor(JLabel label, JComponent component, int key) {
		label.setLabelFor(component);
		label.setDisplayedMnemonic(key);
	}

	/**
	 * @return {@link #okButtonToBeFocussed}, also true, wenn bei Aufruf der Maske
	 *         der OK-Button den Focus haben soll, sonst false.
	 */
	protected boolean isOkButtonToBeFocussed() {
		return okButtonToBeFocussed;
	}

	/**
	 * Die Methode soll die OK-Schaltfläche fokussieren, wenn gewünscht
	 * 
	 * @param okButtonToBeFocussed true, wenn bei Aufruf der Maske der OK-Button den
	 *                             Focus haben soll
	 */
	protected void setOkButtonToBeFocussed(boolean okButtonToBeFocussed) {
		this.okButtonToBeFocussed = okButtonToBeFocussed;
	}

	/**
	 * @return gibt die okAction als AbstractAction zurück.
	 */
	protected AbstractAction getOkAction() {
		return okAction;
	}

	/**
	 * Die Methode ruft den OK-Button auf
	 * 
	 * @param ae ein ActionEvent
	 */
	protected void performOkAction(ActionEvent ae) {
		okAction.actionPerformed(ae);
	}

	/**
	 * Die Methode aktiviert oder deaktiviert den Ok-Button
	 * 
	 * @param enabled true oder false
	 */
	protected void setOkActionEnabled(boolean enabled) {
		okAction.setEnabled(enabled);
	}

	/**
	 * Setzt für ein {@link JFormattedTextField} die {@link AbstractAction} in die
	 * ActionMap und InputMap, die bei Enter ausgeführt werden soll. Zudem fügt die
	 * Methode das Verhalten hinzu, dass bei FocusErlangung der Inhalt markiert
	 * wird.
	 * 
	 * @param ftf    das {@link JFormattedTextField}, dessen Entertaste belegt
	 *               werden soll
	 * @param action die {@link AbstractAction}, die bei Entertaste ausgeführt
	 *               werden soll
	 */
	protected void setEnterAction(JFormattedTextField ftf, AbstractAction action) {
		if (action != null) {
			ftf.getInputMap(JComponent.WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
					action.getValue(Action.ACTION_COMMAND_KEY));
			ftf.getActionMap().put(action.getValue(Action.ACTION_COMMAND_KEY), action);
		}
		ftf.setHorizontalAlignment(JTextField.RIGHT);
		ftf.addFocusListener(BEI_FOCUS_MARKIEREN);
	}

	/**
	 * Die Methode dient dazu, TODO
	 * 
	 */
	private void fensterSchliessenBeiAbbruch() {
		isOk = false;
		setVisible(false);
		dispose();
	}
}